@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004 - 2009 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,12 +14,14 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <sys/param.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <sys/mman.h>
+#include <sys/resource.h>
 
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -50,6 +52,9 @@
 #include "proto.h"
 #include "extern.h"
 
+/* count in additional fds like the listener sockets */
+#define CON_MAX		CON + 16
+
 /*
  * global vars local
  */
@@ -109,7 +114,7 @@ const char *http_fn_auw		= "WWW-Authenticate:";
 const char *http_fn_ran		= "Range:";
 const char *http_fn_rac		= "Content-Range:";
 const char *http_fn_aen		= "Accept-Encoding:";
-const char *http_fv_srv		= "nostromo 1.9.3";
+const char *http_fv_srv		= "nostromo 1.9.5";
 const char *http_fv_pr0	=	 "HTTP/1.0";
 const char *http_fv_pr1	=	 "HTTP/1.1";
 const char *http_fv_cgi		= "CGI/1.1";
@@ -126,7 +131,7 @@ char	http_sig[128];
 char	http_path[128];
 
 struct cfg		config;
-struct connection	c[CON];
+struct connection	c[CON_MAX];
 struct listener		l[LST + 1];
 
 /*
@@ -441,8 +446,9 @@ main(int argc, char *argv[])
 	struct in_addr		ina;
 	struct in6_addr		ina6;
 	struct timeval		tv;
+	struct rlimit		rlp;
 	socklen_t		ssize;
-	fd_set			master_r, master_w, set_r, set_w;
+	fd_set			*master_r, *master_w, *set_r, *set_w;
 	time_t			tnow;
 	SSL_CTX			*ssl_ctx = NULL;
 	FILE			*file;
@@ -727,13 +733,36 @@ main(int argc, char *argv[])
 			err(1, "SSL_CTX_check_private_key");
 	}
 
+	/*
+	 * calculate maximal concurrent connections
+	 */
+	sdlim = CON_MAX;
+
+	rlp.rlim_cur = sdlim;
+	rlp.rlim_max = sdlim;
+	r = setrlimit(RLIMIT_NOFILE, &rlp);
+	if (r == -1)
+		err(1, "setrlimit");
+
 	/*
 	 * initialize select sets
 	 */
-	FD_ZERO(&master_r);
-	FD_ZERO(&master_w);
-	FD_ZERO(&set_r);
-	FD_ZERO(&set_w);
+	master_r = (fd_set *)calloc(howmany(sdlim + 1, NFDBITS),
+	    sizeof(fd_mask));
+	if (master_r == NULL)
+		err(1, "calloc");
+	master_w = (fd_set *)calloc(howmany(sdlim + 1, NFDBITS),
+	    sizeof(fd_mask));
+	if (master_w == NULL)
+		err(1, "calloc");
+	set_r = (fd_set *)calloc(howmany(sdlim + 1, NFDBITS),
+	    sizeof(fd_mask));
+	if (set_r == NULL)
+		err(1, "calloc");
+	set_w = (fd_set *)calloc(howmany(sdlim + 1, NFDBITS),
+	    sizeof(fd_mask));
+	if (set_w == NULL)
+		err(1, "calloc");
 
 	i = 0;
 	/*
@@ -789,7 +818,7 @@ main(int argc, char *argv[])
 				err(1, "listen");
 
 			/* add listener socket to master set */
-			FD_SET(l[i].sd, &master_r);
+			FD_SET(l[i].sd, master_r);
 
 			/* keep track of the biggest socket descriptor */
 			sdmax = l[i].sd;
@@ -858,7 +887,7 @@ main(int argc, char *argv[])
 				err(1, "listen");
 
 			/* add ssl listener socket to master set */
-			FD_SET(l[i].sd, &master_r);
+			FD_SET(l[i].sd, master_r);
 
 			/* keep track of the biggest socket descriptor */
 			sdmax = l[i].sd;
@@ -933,7 +962,7 @@ main(int argc, char *argv[])
 				err(1, "listen");
 
 			/* add ipv6 listener to master set */
-			FD_SET(l[i].sd, &master_r);
+			FD_SET(l[i].sd, master_r);
 
 			/* keep track of the biggest socket descriptor */
 			sdmax = l[i].sd;
@@ -1007,7 +1036,7 @@ main(int argc, char *argv[])
 				err(1, "listen");
 
 			/* add ipv6 listener to master set */
-			FD_SET(l[i].sd, &master_r);
+			FD_SET(l[i].sd, master_r);
 
 			/* keep track of the biggest socket descriptor */
 			sdmax = l[i].sd;
@@ -1031,11 +1060,6 @@ main(int argc, char *argv[])
 		exit(1);
 	}
 
-	/*
-	 * calculate maximal concurrent connections
-	 */
-	sdlim = CON + sdmax;
-
 	/*
 	 * signal handler
 	 */
@@ -1148,10 +1172,21 @@ main(int argc, char *argv[])
 		}
 	}
 
+	/*
+	 * recheck our fd limits
+	 */
+	r = getrlimit(RLIMIT_NOFILE, &rlp);
+	if (r == -1) {
+		syslog(LOG_ERR, "getrlimit: %s", strerror(errno));
+		exit(1);
+	}
+
 	/*
 	 * nostromo ready
 	 */
 	syslog(LOG_INFO, "started");
+	syslog(LOG_INFO, "max. file descriptors = %d (cur) / %d (max)",
+	    (int)rlp.rlim_cur, (int)rlp.rlim_max);
 	if (config.ssl)
 		syslog(LOG_INFO, "SSL enabled on port %d", config.sslport);
 	if (opt_r)
@@ -1168,13 +1203,13 @@ main(int argc, char *argv[])
 	 */
 	while (!quit) {
 		/* copy master set */
-		set_r = master_r;
-		set_w = master_w;
+		memcpy(set_r, master_r, howmany(sdlim + 1, NFDBITS));
+		memcpy(set_w, master_w, howmany(sdlim + 1, NFDBITS));
 
 		/* select is waiting for some action */
 		tv.tv_sec = TON;
 		tv.tv_usec = 0;
-		if (select(sdmax + 1, &set_r, &set_w, NULL, &tv) == -1) {
+		if (select(sdmax + 1, set_r, set_w, NULL, &tv) == -1) {
 			if (errno == EINTR) {
 				usleep(1);
 				continue;
@@ -1193,8 +1228,8 @@ main(int argc, char *argv[])
 				continue;
 
 			/* check connection read and write state */
-			rd = FD_ISSET(sdnow, &set_r);
-			wr = FD_ISSET(sdnow, &set_w);
+			rd = FD_ISSET(sdnow, set_r);
+			wr = FD_ISSET(sdnow, set_w);
 
 			/*
 			 * existing connection idles, check for timeout
@@ -1278,7 +1313,7 @@ main(int argc, char *argv[])
 				}
 
 				/* check connection limit */
-				if (sdnew > sdlim) {
+				if (sdnew == (sdlim - 1)) {
 					h = http_head(http_s_503, "-",
 					    c[sdnew].ip, 0);
 					b = http_body(http_s_503, "", h, 0);
@@ -1290,11 +1325,11 @@ main(int argc, char *argv[])
 					c[sdnew].x_sta = 0;
 					free(h);
 					free(b);
-					FD_SET(sdnew, &master_w);
+					FD_SET(sdnew, master_w);
 				}
 
 				/* add new connection to read fdset */
-				FD_SET(sdnew, &master_r);
+				FD_SET(sdnew, master_r);
 
 				/* set highest socket */
 				if (sdnew > sdmax)
@@ -1346,7 +1381,7 @@ main(int argc, char *argv[])
 				}
 				/* all partial files sent */
 				if (c[sdnow].pfdo == 0) {
-					FD_CLR(sdnow, &master_w);
+					FD_CLR(sdnow, master_w);
 					if (c[sdnow].x_sta == 0)
 						goto quit;
 				}
@@ -1465,8 +1500,8 @@ main(int argc, char *argv[])
 			for (i = 0; i < k; i++) {
 				s = 0;
 				r = 1;
-				s = http_verify(header[i], c[sdnow].ip, sdnow,
-				    i);
+				s = http_verify(header[i], sizeof(header[i]),
+				    c[sdnow].ip, sdnow, i);
 				if (s == 1)
 					r = http_proc(header[i], body, i, size,
 					    sdnow);
@@ -1491,7 +1526,7 @@ main(int argc, char *argv[])
 
 			/* connection has open files, add to write fdset */
 			if (c[sdnow].pfdo > 0)
-				FD_SET(sdnow, &master_w);
+				FD_SET(sdnow, master_w);
 
 			/* check keep-alive */
 			if (r != 1) {
@@ -1504,8 +1539,8 @@ main(int argc, char *argv[])
 				free(c[sdnow].rbuf);
 				free(c[sdnow].wbuf);
 				close(sdnow);
-				FD_CLR(sdnow, &master_r);
-				FD_CLR(sdnow, &master_w);
+				FD_CLR(sdnow, master_r);
+				FD_CLR(sdnow, master_w);
 			}
 		}
 	}
