@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004 - 2009 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -27,7 +27,9 @@ int		sdlisten(int, int);
 /*
  * http.c
  */
-int		http_verify(const char *, const char *, const int, const int);
+int		http_decode_header_uri(char *, const int);
+int		http_verify(char *, const int, const char *, const int,
+		    const int);
 int		http_proc(const char *, char *, const int, const int,
 		    const int);
 int		http_cgi_getexec(char *, char *, const char *, const int,
